# -*- coding: utf-8 -*-

# Copyright 2008-2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import math
from skoollib import Game, Udg, MAIN_CODE_ID
from skoolhtml import HtmlWriter, FileInfo, Table, Header, Cell, join

class MicrosphereGame(Game):
    def _calculate_tap_index(self, tap_address_table, max_tap):
        min_tap = tap_address_table % 256
        self.tap_addresses = {}
        self.tap_descs = {}
        for t in range(min_tap, max_tap + 1, 2):
            address = tap_address_table + t - min_tap
            tap_address = self.snapshot[address] + 256 * self.snapshot[address + 1]
            tap_entry = self.entries.get(tap_address)
            if tap_entry:
                sep_index = tap_entry.description.index(':')
                self.tap_descs[t] = tap_entry.description[sep_index + 1:].strip()
                self.tap_addresses[t] = tap_address

    def _flip_byte(self, byte):
        f = 0
        for b in range(8):
            f = 2 * f + (byte & 1)
            byte /= 2
        return f

    def _flip_udg(self, udg):
        for index in range(8):
            udg.data[index] = self._flip_byte(udg.data[index])
            udg.mask[index] = self._flip_byte(udg.mask[index])

    def _get_sprite_udg(self, state, attr, ref_page, udg_page):
        ref_addr = state + 256 * ref_page
        ref = self.snapshot[ref_addr]
        if ref == 0:
            udg = [0] * 8
            mask = [255] * 8
        else:
            udg_addr = ref + 256 * udg_page
            udg = self.snapshot[udg_addr:udg_addr + 4096:512]
            mask = self.snapshot[udg_addr + 256:udg_addr + 4352:512]
        return Udg(attr, udg, mask, ref_addr=ref_addr, ref=ref, udg_page=udg_page)

    def _get_sprite_size(self, state):
        return 3, 4

    def build_sprite(self, state, attr=None, udg_page=None):
        width, height = self._get_sprite_size(state)
        udg_array = []
        for row in range(height):
            udg_array.append([self.get_sprite_udg(state, attr, row, col, udg_page) for col in range(width)])
        return udg_array

    def get_font_bitmap(self, character):
        base_address = ord(character) + self.font_address + 256
        width = self.snapshot[base_address - 256]
        return self.snapshot[base_address:base_address + 256 * width:256]

    def get_text(self, words):
        columns = []
        for character in words:
            columns.append(0)
            columns.extend(self.get_font_bitmap(character))
        udg_array = []
        for col in range(len(columns)):
            col_byte = columns[col]
            udg_index = col // 8
            bit = int(math.pow(2, 7 - (col - 8 * udg_index)))
            if udg_index == len(udg_array):
                udg_array.append([0] * 8)
            udg = udg_array[udg_index]
            for b in range(8):
                udg[7 - b] |= bit * (col_byte & 1)
                col_byte /= 2
        return udg_array

    def _draw_text(self, udgs, words, x, y):
        text_udgs = self.get_text(words)
        for col in range(len(text_udgs)):
            udg = udgs[y][col + x].data
            for index in range(8):
                udg[index] |= text_udgs[col][index]

    def get_skool_udgs(self, x, y, w, h, show_chars=False, show_x=0):
        skool_udgs = []
        for row in range(y, y + h):
            skool_udgs.append([])
            for col in range(x, x + w):
                skool_udgs[-1].append(self.get_skool_udg(row, col, show_chars))
        if show_x > 0:
            line = [Udg(8 * (7 - skool_x % 2), [0] * 8) for skool_x in range(x, x + w)]
            for skool_x in range(x, x + w):
                if skool_x % show_x == 0:
                    self._draw_text([line], str(skool_x), skool_x - x, 0)
            skool_udgs.insert(0, line)
            skool_udgs.append(line)
        return skool_udgs

    def _superimpose_sprite_udgs(self, skool_udg, x, y):
        for state, char_x, char_y in self.get_initial_states():
            if char_y <= y <= char_y + 3 and char_x <= x <= char_x + 2:
                sprite_udg = self.get_sprite_udg(state, None, y - char_y, x - char_x)
                for i in range(8):
                    skool_udg.data[i] = (skool_udg.data[i] | sprite_udg.data[i]) & sprite_udg.mask[i]

    def _get_char_buf_descs(self):
        char_buf_descs = []
        for _, bytes, descs in self.get_ref_items('Data:CharBuf'):
            char_buf_descs.append((bytes, descs))
        return char_buf_descs

class MicrosphereHtmlWriter(HtmlWriter):
    def _get_paths(self):
        paths = HtmlWriter._get_paths(self)
        paths['LessonsPath'] = 'lessons'
        return paths

    def _get_links(self):
        links = HtmlWriter._get_links(self)
        links['CommandLists'] = ('Command lists', '')
        links['MainTimetable'] = ('Main timetable', '')
        return links

    def as_img(self, cwd, num, scale=2, mask=1, attr=120, udg_page=None, fname_suffix=''):
        mask_infix = 'm' if mask else 'u'
        udg_page_infix = udg_page if udg_page is not None else ''
        snapshot_name = self.skool.get_snapshot_name()
        snapshot_infix = '_%s' % snapshot_name if snapshot_name else ''
        fname = 'as%03i_%ix%i%s%s%s%s' % (num & 255, attr, scale, mask_infix, udg_page_infix, snapshot_infix, fname_suffix)
        asimg_fname = join('images', 'animatory_states', '%s.png' % fname)
        if self.need_image(asimg_fname):
            self.write_image(self.skool.build_sprite(num, attr, udg_page), asimg_fname, scale=scale, mask=mask)
        img_path = FileInfo.relpath(cwd, asimg_fname)
        return '<img alt="Animatory state %i" src="%s" />' % (num, img_path)

    def play_area(self, cwd, fname, x, y, w=1, h=1, scale=2, show_chars=0, show_x=0):
        img_fname = self.get_img_fname(fname, 'images/play_area')
        if self.need_image(img_fname):
            self.write_image(self.skool.get_skool_udgs(x, y, w, h, show_chars, show_x), img_fname, scale=scale)
        img_path = FileInfo.relpath(cwd, img_fname)
        return '<img alt="%s" src="%s" />' % (fname, img_path)

    def write_playarea_tile_table(self, cwd, tiles, id_prefix='s'):
        lines = []
        lines.append('<table class="tiles">\n')
        for row in tiles:
            lines.append('<tr>\n')
            for tile in row:
                lines.append('<td>')
                if tile:
                    bubble_id = '%s%ix%i' % (id_prefix, tile.x, tile.y)
                    title = '(x,y)=(%i,%i)' % (tile.x, tile.y)
                    img_fname = '%x%x.png' % (tile.ref + 256 * tile.udg_page, tile.attr)
                    img_path = join(cwd, img_fname)
                    if self.need_image(img_path):
                        self.write_image([[tile]], img_path, scale=4)
                    img = '<img alt="%i,%i" src="%s"/>' % (tile.x, tile.y, img_fname)
                    lines.append('<div class="T" onclick="s(\'%s\')">%s</div>' % (bubble_id, img))
                    lines += self._get_bubble(bubble_id, img, title, '<br/>'.join(self.get_playarea_tile_text(tile)))
                lines.append('</td>\n')
            lines.append('</tr>\n')
        lines.append('</table>\n')
        return lines

    def write_mutable_tile_tables(self, cwd):
        return []

    def play_area_tiles(self, cwd, w, h):
        lines = []
        lines += self.write_playarea_tile_table(cwd, self.skool.get_skool_udgs(0, 0, w, h))
        lines += self.write_mutable_tile_tables(cwd)
        return ''.join(lines)

    def _animatory_state_row(self, cwd, rownum):
        lines = ['<tr>']
        for state in (rownum, rownum + 128):
            lines.append('<td>')
            lines.append('<a name="%i" />%s' % (state, self.as_img(cwd, state)))
            lines.append('</td>')
        lines.append('<td>%i, %i: %s</td>' % (rownum, rownum + 128, self.skool.as_descs.get(rownum, '-')))
        lines.append('</tr>')
        return lines

    def animatory_states(self, cwd):
        lines = ['<table>']
        for n in range(128):
            lines += self._animatory_state_row(cwd, n)
        lines.append('</table>')
        return '\n'.join(lines)

    def _get_animatory_state_tiles_row(self, n):
        states = ((n, None),)
        states_desc = '%i: %s' % (n, self.skool.as_descs[n])
        return [(states, states_desc)]

    def _get_bubble(self, bubble_id, img, title, text):
        return [
            '<div id="%s" class="B P" onclick="h(this)">' % bubble_id,
            '<div class="H">%s</div>' % title,
            '<table><tr><td>%s</td><td>%s</td></tr></table>' % (img, text),
            '</div>'
        ]

    def _get_sprite_tile_img_fname(self, tile, state):
        return '%x.png' % (tile.ref + 256 * tile.udg_page)

    def astiles(self, cwd):
        lines = []
        attr = 120
        lines.append('<table>\n')
        for n in range(128):
            row = []
            for state_specs, states_desc in self._get_animatory_state_tiles_row(n):
                lines.append('<tr>\n')
                for state, udg_page in state_specs:
                    lines.append('<td>\n')
                    lines.append('<table class="A">\n')
                    sprite = self.skool.build_sprite(state, attr, udg_page)
                    num_rows = len(sprite)
                    for row_num in range(num_rows):
                        lines.append('<tr>\n')
                        row = sprite[row_num]
                        for col_num in range(len(row)):
                            tile = row[col_num]
                            bubble_id_suffix = '%02x' % udg_page if udg_page is not None else ''
                            bubble_id = 'B%02x%x%s' % (state, row_num + num_rows * col_num, bubble_id_suffix)
                            title = '%i: row %i, column %i' % (state, row_num, col_num)
                            img_fname = self._get_sprite_tile_img_fname(tile, state)
                            img_path = join(cwd, img_fname)
                            if self.need_image(img_path):
                                self.write_image([[tile]], img_path, scale=4, mask=1)
                            img = '<img alt="%i:%i,%i" src="%s"/>' % (state, row_num, col_num, img_fname)
                            text = ['<i>lsb</i>=%i, <i>ref_page</i>=%i' % (tile.ref_addr % 256, tile.ref_addr // 256)]
                            text.append('<i>ref_addr</i>=(<i>lsb</i>,<i>ref_page</i>)=%i' % tile.ref_addr)
                            if tile.ref == 0:
                                text.append('<i>ref</i>=%i' % tile.ref)
                            else:
                                text.append('<i>ref</i>=%i, <i>udg_page</i>=%i' % (tile.ref, tile.udg_page))
                                text.append('<i>udg_addr</i>=(<i>ref</i>,<i>udg_page</i>)=%i' % (tile.ref + 256 * tile.udg_page))
                            lines.append('<td>')
                            lines.append('<div class="T" onclick="s(\'%s\')">%s</div>' % (bubble_id, img))
                            lines += self._get_bubble(bubble_id, img, title, '<br/>'.join(text))
                            lines.append('</td>\n')
                        lines.append('</tr>\n')
                    lines.append('</table>\n')
                    lines.append('</td>\n')
                lines.append('<td>%s</td>\n' % states_desc)
                lines.append('</tr>\n')
        lines.append('</table>\n')
        return ''.join(lines)

    def _get_pt_link(self, cwd, char_num):
        pt_address = self.skool.min_lesson + 256 * char_num
        pt_file = FileInfo.asm_relpath(cwd, pt_address, self.skool.get_code_path(MAIN_CODE_ID))
        return '<a class="link" href="%s">%i</a>' % (pt_file, char_num)

    def lesson_index(self, cwd):
        li_table = Table('lesson', 'ID', 'Lesson')
        for i in range(self.skool.min_lesson, self.skool.max_lesson + 1):
            row = []
            lesson_file = FileInfo.relpath(cwd, self.paths['Lesson%i' % i])
            row.append(Cell('<a class="link" href="%s">%i</a>' % (lesson_file, i), "lessonType"))
            row.append(Cell(self.skool.lesson_descs[i], "lesson"))
            li_table.add_row(row)
        return li_table.html()

    def _lesson_prev_next(self, prev, next, up):
        prev_text = next_text = up_text = ''
        if prev:
            prev_text = 'Prev: %s' % prev
        if next:
            next_text = 'Next: %s' % next
        if up:
            up_text = 'Up: %s' % up
        lines = ['<table class="prevNext">', '<tr>']
        for td_class, text in (('prev', prev_text), ('up', up_text), ('next', next_text)):
            if text:
                lines.append('<td class="%s">%s</td>' % (td_class, text))
            else:
                lines.append('<td class="%s"/>' % td_class)
        lines.append('</tr>')
        lines.append('</table>')
        return lines

    def lesson(self, cwd, lesson):
        lines = []
        index_path = FileInfo.relpath(cwd, self.paths['LessonIndex'])

        prev = None if lesson == self.skool.min_lesson else self.expand('#LINK:Lesson%i(%i)' % (lesson - 1, lesson - 1), cwd)
        up = '<a class="link" href="%s">Index</a>' % index_path
        next = None if lesson == self.skool.max_lesson else self.expand('#LINK:Lesson%i(%i)' % (lesson + 1, lesson + 1), cwd)
        prev_up_next = self._lesson_prev_next(prev, next, up)

        lines += prev_up_next
        lines.append('<div class="description">Lesson %i: %s</div>' % (lesson, self.skool.lesson_descs[lesson]))
        lines.append('<table class="lesson">')
        lines.append('<tr>')
        lines.append('<th>Character</th>')
        lines.append('<th>Address</th>')
        lines.append('<th>Command list</th>')
        lines.append('</tr>')
        for c in range(self.skool.min_character, self.skool.max_character + 1):
            lesson_address = lesson + 256 * c
            tap = self.skool.snapshot[lesson_address]
            pt_link = self._get_pt_link(cwd, c)
            char_desc = self.skool.characters[c]
            tap_address = self.skool.tap_addresses[tap]
            tap_link = self.expand('#R%i(%i)' % (tap_address, tap), cwd)
            tap_desc = self.skool.tap_descs[tap]
            lines.append('<tr>')
            lines.append('<td class="lessonCharacter">%s - %s</td>' % (pt_link, char_desc))
            lines.append('<td class="lessonAddress">%i</td>' % lesson_address)
            lines.append('<td class="lesson">%s - %s</td>' % (tap_link, tap_desc))
            lines.append('</tr>')
        lines.append('</table>')
        lines += prev_up_next

        return '\n'.join(lines)

    def personal_timetables(self, cwd):
        pt_table = Table('timetable', 'No.', 'Character')
        for c in range(self.skool.min_character, self.skool.max_character + 1):
            row = []
            row.append(Cell(self._get_pt_link(cwd, c)))
            row.append(Cell(self.skool.characters[c]))
            pt_table.add_row(row)
        return pt_table.html()

    def _characters_table(self, cwd, char_buffer_base_byte):
        characters = self.skool.characters
        if not characters:
            return ''
        first_char_num = min(characters.keys())

        cols = 3

        headers = []
        for c in range(cols):
            headers.append(Header('Page'))
            headers.append(Header('Address'))
            headers.append(Header('Character'))
            if c < cols - 1:
                headers.append(Cell('', "transparent"))
        char_table = Table('cbuffer')
        char_table.add_row(headers)

        rows = (cols - 1  + len(characters)) // cols
        for r in range(rows):
            row = []
            for c in range(cols):
                char_index = r + c * rows
                if char_index < len(characters):
                    char_num = first_char_num + char_index
                    row.append(Cell(char_num, "centre"))
                    buf_addr = char_buffer_base_byte + char_num * 256
                    entry_link = self.expand('#R%i' % buf_addr, cwd)
                    row.append(Cell(entry_link, "centre"))
                    row.append(Cell(characters[char_num]))
                    if c < cols - 1:
                        row.append(Cell('', "transparent"))
            char_table.add_row(row)

        return char_table.html()

    def character_buffers(self, cwd, char_buffer_base_byte):
        lines = [self._characters_table(cwd, char_buffer_base_byte)]
        lines.append('<p>The bytes in each character buffer are used as follows:</p>\n')
        cb_table = Table('cbuffer', 'Byte(s)', 'Contents')
        for bytes, entries in self.skool.char_buf_descs:
            rows = len(entries)
            anchors = ''.join(['<a name="%s" />' % name for name in bytes.split(',')])
            row = []
            row.append(Cell(bytes, "centre", rowspan=rows))
            row.append(Cell("%s%s" % (anchors, self.expand(entries[0], cwd)), "byteDesc"))
            cb_table.add_row(row)
            for i in range(1, rows):
                cb_table.add_row((Cell(self.expand(entries[i], cwd), "byteDesc"),))
        lines.append(cb_table.html())
        return ''.join(lines)

class Udg:
    def __init__(self, attr, data, mask=None, attr_addr=None, q_addr=None, ref_addr=None, ref=None, udg_page=None, x=None, y=None):
        self.attr = attr
        self.data = data
        self.mask = mask
        self.attr_addr = attr_addr
        self.q_addr = q_addr
        self.ref_addr = ref_addr
        # We store the UDG reference now in case the snapshot changes before
        # the reference is looked up via self.snapshot[udg.ref_addr] (e.g.
        # because a saved snapshot is restored after
        # BackToSkool._alter_skool_udgs() has been called)
        self.ref = ref
        self.udg_page = udg_page
        self.x = x
        self.y = y
