# -*- coding: utf-8 -*-

# Copyright 2008-2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import os

# The ID of the main disassembly
MAIN_CODE_ID = 'main'

class Game:
    def __init__(self, skool_parser, ref_parser, code_id=MAIN_CODE_ID):
        self.parser = skool_parser
        self.snapshot = self.parser.snapshot
        self._snapshots = [(self.snapshot, '')]
        self.entries = self.parser.entries
        self.memory_map = [e for e in self.parser.memory_map if e.ctl != 'i']

        self.ref_parser = ref_parser
        self.info = self.ref_parser.build_dictionary('Info')

        self.paths = self.ref_parser.build_dictionary('Paths')
        self.titles = self.ref_parser.build_dictionary('Titles')
        link_text = self.ref_parser.build_dictionary('Links')
        self.page_ids = []
        self.pages = {}
        for page_id, details in self.ref_parser.build_dictionaries('Page'):
            self.page_ids.append(page_id)
            self.pages[page_id] = details
        self.page_contents = {}
        for page_id, page in self.pages.items():
            self.page_contents[page_id] = page.get('PageContent', self.ref_parser.get_section('PageContent:%s' % page_id))
            path = page.get('Content')
            if path:
                self.page_ids.remove(page_id)
            else:
                path = page['Path']
            self.paths[page_id] = path
            self.titles[page_id] = page.get('Title', page_id)
            link_text[page_id] = page.get('Link', self.titles[page_id])
        self.links = self._parse_links(link_text)

        self.other_code = self.ref_parser.build_dictionaries('OtherCode')
        self.code_path = self.get_code_path(code_id)
        self.game_vars = self.ref_parser.build_dictionary('Game')
        self.colours = self._get_colours(ref_parser.build_dictionary('Colours'))
        def_game_name = os.path.basename(self.parser.skoolfile)
        suffix = '.skool'
        if def_game_name.endswith(suffix):
            def_game_name = def_game_name[:-len(suffix)]
        self.game = self.game_vars.get('Game', def_game_name)
        self.map_details = self.ref_parser.build_dictionary('MapDetails')

    def _get_colours(self, cdict):
        colours = {}
        for k, v in cdict.items():
            colours[k] = tuple([int(n) for n in v.split(',')])
        return colours

    def _parse_links(self, links):
        new_links = {}
        for page_id, link_text in links.items():
            if link_text.startswith('['):
                sb_count = 1
                index = 1
                length = len(link_text)
                while index < length and sb_count > 0:
                    if link_text[index] == '[':
                        sb_count += 1
                    elif link_text[index] == ']':
                        sb_count -= 1
                    index += 1
                if sb_count == 0:
                    new_links[page_id] = (link_text[1:index - 1], link_text[index:])
            if page_id not in new_links:
                new_links[page_id] = (link_text.strip(), '')
        return new_links

    def get_code_path(self, code_id):
        if code_id == MAIN_CODE_ID:
            return self.paths.get('CodePath', 'asm')
        for c_id, code in self.other_code:
            if c_id == code_id:
                return code['Path']

    def has_gbuffer(self):
        for entry in self.memory_map:
            if entry.ctl == 'g':
                return True
        return False

    def get_data_dictionary(self, data_name):
        return self.ref_parser.get_data_dictionary(data_name)

    def has_ref_items(self, section_type):
        return self.ref_parser.has_ref_items(section_type)

    def get_section(self, section_name):
        return self.ref_parser.get_section(section_name)

    def get_ref_items(self, section_type, keep_lines=False, as_list=False):
        return self.ref_parser.get_ref_items(section_type, keep_lines, as_list)

    def get_list(self, section_name, default=()):
        return self.ref_parser.get_list(section_name) or default

    def get_entry(self, address):
        return self.parser.get_entry(address)

    def get_entry_point_refs(self, address):
        return self.parser.get_entry_point_refs(address)

    def get_container(self, address):
        return self.parser.get_container(address)

    def get_instruction_addr_str(self, address):
        return self.parser.instructions[address][0].addr_str

    def get_snapshot_name(self):
        return self._snapshots[-1][1]

    def pop_snapshot(self):
        self.snapshot = self._snapshots.pop()[0]

    def push_snapshot(self, name=''):
        self._snapshots.append((self.snapshot[:], name))

    def _get_screen_udg(self, row, col):
        attr = self.snapshot[22528 + 32 * row + col]
        address = 16384 + 2048 * (row // 8) + 32 * (row % 8) + col
        return Udg(attr, self.snapshot[address:address + 2048:256])

    def get_screenshot(self, x=0, y=0, w=32, h=24):
        width = min((w, 32 - x))
        height = min((h, 24 - y))
        scr_udgs = []
        for r in range(y, y + height):
            scr_udgs.append([self._get_screen_udg(r, c) for c in range(x, x + width)])
        return scr_udgs

class Udg:
    def __init__(self, attr, data, mask=None):
        self.attr = attr
        self.data = data
        self.mask = mask
