# -*- coding: utf-8 -*-
import os
from shutil import rmtree
import unittest
from skoolkittest import SkoolKitTestCase

ODIR = 'html'
TEST_REF = 'test-html.ref'
TEST_NO_REF = 'test-html-no-ref.skool'

USAGE = """Usage: skool2html.py [options] FILE [FILE...]

  Convert skool files and ref files to HTML. FILE may be a regular file, or '-'
  for standard input.

Options:
  -V       Show SkoolKit version number and exit
  -q       Be quiet
  -d DIR   Write files in this directory
  -o       Overwrite existing image files
  -i LIB   Use this image library ('gd' or 'PIL'; default is 'gd')
  -l       Write disassembly in lower case
  -u       Write disassembly in upper case
  -D       Write disassembly in decimal
  -H       Write disassembly in hexadecimal
  -P LIST  Write only these custom pages (when '-wP' is specified)
  -wX      Write only these files, where X is one or more of:
             B = Graphic glitches
             b = Bugs
             d = Disassembly files
             G = Game status buffer
             g = Graphics
             i = Disassembly index
             m = Memory maps
             o = Other code
             P = Pages defined in the ref file
             p = Pokes
             t = Trivia
             y = Glossary"""

OUTPUT_NO_REF = """Creating directory html
Found no ref file for test-html-no-ref.skool
Parsing test-html-no-ref.skool
Creating directory html/test-html-no-ref
Copying style.css to html/test-html-no-ref/style.css
  Writing disassembly files in test-html-no-ref/asm
  Writing test-html-no-ref/maps/all.html
  Writing test-html-no-ref/maps/routines.html
  Writing test-html-no-ref/maps/data.html
  Writing test-html-no-ref/maps/messages.html
  Writing test-html-no-ref/index.html"""

class Skool2HtmlTest(SkoolKitTestCase):
    def setUp(self):
        rmtree(ODIR, True)

    def tearDown(self):
        rmtree(ODIR, True)

    def test_no_arguments(self):
        cmdline, output, error = self.run_skool2html(err_lines=True)
        self.assert_output_equal(error, USAGE.split('\n'), False)

    def test_PIL(self):
        cmdline, output, error = self.run_skool2html('-i PIL -wi -d %s %s' % (ODIR, TEST_NO_REF))
        self.assert_no_errors(error, cmdline)
        self.assertEqual(output[0], "Using 'PIL' image library")
        self.assertEqual(output[-1][:37], "  Writing test-html-no-ref/index.html")

    def test_no_ref(self):
        cmdline, output, error = self.run_skool2html('-d %s %s' % (ODIR, TEST_NO_REF))
        self.assert_no_errors(error, cmdline)
        self.assert_output_equal(output, OUTPUT_NO_REF.split('\n'))

    def test_nonexistent_file(self):
        skoolfile = 'xyz.skool'
        cmdline, output, error = self.run_skool2html('-d %s %s' % (ODIR, skoolfile))
        self.assertEqual(error.strip(), 'ERROR: %s: skool file not found' % skoolfile)
        reffile = 'zyx.ref'
        cmdline, output, error = self.run_skool2html('-d %s %s' % (ODIR, reffile))
        self.assertEqual(error.strip(), 'ERROR: %s: ref file not found' % reffile)

    def test_invalid_page_id(self):
        page_id = 'NonexistentPage'
        cmdline, output, error = self.run_skool2html('-d %s -wP -P %s %s' % (ODIR, page_id, TEST_REF))
        self.assertEqual(error.strip(), 'ERROR: Invalid page ID: %s' % page_id)

if __name__ == '__main__':
    unittest.main()
